print("import: 'numcodecs'")
import numcodecs

print("import: 'numcodecs.blosc'")
import numcodecs.blosc

print("import: 'numcodecs.lz4'")
import numcodecs.lz4

print("import: 'numcodecs.zstd'")
import numcodecs.zstd

